from datetime import datetime
import pytz

def timezone_converter(timestamp, source_tz, destination_tz, fmt=None):
    """
    Converts timestamp from a source timezone to a destination timezone. 
    To see commonly used timezone options run: 
    
    ```from pytz import common_timezones```
    
    :param timestamp: (datetime.datetime) timestamp to convert
    :param source_tz: (pytz.timezone) source timezone to convert 
    :param destination_tz: (pytz.timezone) destination timezone
    :param fmt: (str) optional - timestamp format to pass to strftime
    
    :returns: (datetime.datetime) converted timestamp
    """
    dest_obj = pytz.timezone(destination_tz)
    source_obj = pytz.timezone(source_tz).localize(timestamp.replace(tzinfo=None))
    converted = dest_obj.normalize(source_obj.astimezone(dest_obj))
    return converted if fmt is None else converted.strftime(fmt)


def current_timestamp(tz='UTC', fmt=None):
    """
    Returns current timestamp in desired timezone (per pytz nomenclature)

    ```from pytz import common_timezones```

    :param tz: (pytz.timezone) desired timezone
    :param fmt: (str) optional - timestamp format to pass to strftime
    :returns: (datetime.datetime) timestamp
    """
    return timezone_converter(datetime.utcnow(), 'UTC', tz, fmt=fmt)


def current_timestamp_for_mysql(tz='UTC', fmt="%Y-%m-%d_%H:%M:%S"):
    """
    Returns current timestamp in desired timezone (per pytz nomenclature)

    ```from pytz import common_timezones```

    :param tz: (pytz.timezone) desired timezone
    :param fmt: (str) optional - timestamp format to pass to strftime, defaults to mysql compatible format
    :returns: (datetime.datetime) timestamp
    """
    return timezone_converter(datetime.utcnow(), 'UTC', tz, fmt=fmt)


"""
Methods for checking installed and latest versions of microns packages.
"""

import traceback
import requests
try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata
import logging
import re
import sys
import json
from pathlib import Path
from .filepath_utils import find_all_matching_files


def parse_version(text: str):
    """
    Extracts __version__ from raw text if __version__ follows semantic versioning (https://semver.org/).
    
    Also compatible with a direct semantic version input, i.e.: "x.y.z". 
    
    :param text (str): the text containing the version.
    :returns (str): version if parsed successfully else ""
    """
    semver = "^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
    version_search = re.search('__version__.*', text)
    text = version_search.group() if version_search is not None else text
    text = text.split('=')[1].strip(' "'" '") if len(text.split('='))>1 else text.strip(' "'" '")
    parsed = re.search(semver, text)
    return parsed.group() if parsed else ""


def check_latest_version_from_github(owner, repo, source, branch='main', path_to_version_file=None, warn=True):
    """
    Checks github for the latest version of package.

    :param owner (str): Owner of repository
    :param repo (str): Name of repository that contains package
    :param source (str): 
        options: 
            "commit" - Gets version of latest commit
            "tag" - Gets version of latest tag
            "release" - Gets version of latest release
    :param branch (str): Branch of repository if source='commit', defaults to 'main'.
    :param path_to_version_file (str): Path to version.py file from top of repo if source = "commit". 
    :param warn (bool): If true, warnings enabled.
    :returns (str): If successful, returns latest version, otherwise returns "".
    """
    latest = ""
    try:
        if source == 'commit':
            assert branch is not None, 'Provide branch if source = "commit".'
            assert path_to_version_file is not None, 'Provide path_to_version_file if source = "commit".'
            f = requests.get(f"https://raw.githubusercontent.com/{owner}/{repo}/{branch}/{path_to_version_file}")
            latest = parse_version(f.text)
            
        elif source == 'tag':
            f = requests.get(f"https://api.github.com/repos/{owner}/{repo}/tags")
            if not f.ok:
                logging.error(f'Could not check Github version because: "{f.reason}".')
                return latest
            latest = parse_version(json.loads(f.text)[0]['name'][1:])
            
        elif source == 'release':
            f = requests.get(f"https://api.github.com/repos/{owner}/{repo}/releases")
            if not f.ok:
                logging.error('Could not check Github version because: "{f.reason}".')
                return latest
            latest = parse_version(json.loads(f.text)[0]['tag_name'][1:])
        
        else:
            raise ValueError(f'source: "{source}" not recognized. Options include: "commit", "tag", "release". ')
    except:
        if warn:
            logging.warning('Failed to check latest version from Github.')
            traceback.print_exc()

    return latest


def latest_github_version_checker(owner, repo):
    """
    Returns a function to check latest version from github.

    :param owner (str): Owner of repository
    :param repo (str): Name of repository that contains package
    """
    def inner(source='tag', branch=None, path_to_version_file=None, warn=True):
        """
        :param source (str): 
            options: 
                "commit" - Gets version of latest commit
                "tag" - Gets version of latest tag
                "release" - Gets version of latest release
        :param branch (str): Branch of repository if source='commit', defaults to 'main'.
        :param path_to_version_file (str): Path to version.py file from top of repo if source = "commit". 
        :param warn (bool): If true, warnings enabled.
        :returns (str): If successful, returns latest version, otherwise returns "".
        """
        return check_latest_version_from_github(owner=owner, repo=repo, source=source, branch=branch, path_to_version_file=path_to_version_file, warn=warn) 
    return inner


def check_package_version_from_distributions(package, warn=True):
    """
    Checks importlib metadata for an installed version of the package. 
    Note: packages installed as editable (i.e. pip install -e) will not be found.

    :param package (str): name of package:
    :returns (str):  If successful, returns version, otherwise returns "".
    """
    version = [dist.version for dist in metadata.distributions() if dist.metadata["Name"] == package]
    if not version:
        if warn:
            logging.warning('Package not found in distributions.')
        return ''
    return version[0]


def check_package_version_from_sys_path(package, path_to_version_file, prefix='', warn=True):
    """
    Checks sys.path for package and returns version from internal version.py file.
    
    :param package (str): name of package. must match at the end of the path string in sys.path.
    :param path_to_version_file (str): path to version.py file relative to package path in sys.path.
    :param prefix (str): path to prepend to package
    :param warn (bool): warnings enabled if True
    :return (str): If successful, returns version, otherwise returns "".
    """
    err_base_str = f'Could not get version for package {package} because '

    paths = [Path(p).joinpath(path_to_version_file) for p in sys.path if re.findall(Path(prefix).joinpath(package).as_posix()+'$', p)]
    
    if len(paths)>1:
        if warn:
            logging.warning(err_base_str + f'{len(paths)} paths containing {package} were found in sys.path. Consider adding a prefix for further specification.')
            [print(p) for p in paths]
        return ''
    
    elif len(paths) == 0:
        if warn:
            logging.warning(err_base_str + f'no paths matching {package} were found in sys.path.')
        return ''
    
    else:
        files = find_all_matching_files('version.py', paths[0])

    if len(files) == 0:
        if warn:
            logging.warning(err_base_str + 'no version.py file was found.')
        return ''

    else:
        with open(files[0]) as f:
            lines = f.readlines()[0]
        
    return parse_version(lines)


def check_package_version(package, prefix='', check_if_latest=False, check_if_latest_kwargs={}, warn=True):
    """
    Checks package version.

    :param package (str): name of package (contains setup.py)
    :param prefix (str): path to prepend to package for check_package_version_from_sys_path
    :param check_if_latest (bool): if True, checks if installed version matches latest version on Github 
    :check_if_latest_kwargs (dict): kwargs to pass to :func:`~version_utils.check_latest_version_from_github`
    :param warn (bool): warnings enabled if True
    :returns (str): current package version
    """
    # check installed distributions for versions
    dist_version = check_package_version_from_distributions(package=package, warn=False)

    if not dist_version:
        # check sys.path for versions
        sys_version = check_package_version_from_sys_path(package=package, prefix=prefix, path_to_version_file='..', warn=warn)
        if not sys_version:
            return ''
        else: 
            __version__ = sys_version
    else:
        __version__ = dist_version

    if check_if_latest:
        # check if package version is latest
        latest = check_latest_version_from_github(**check_if_latest_kwargs)

        if __version__ != latest:
            if warn:
                logging.warning(f'You are using {package} version {__version__}, which does not match the latest version on Github, {latest}.')
    
    return __version__


"""
Utils for working with filepaths.
"""
import os
from pathlib import Path
import datetime
import logging
from .datetime_utils import timezone_converter

def find_all_matching_files(name, path):
    """
    Finds all files matching filename within path.

    :param name (str): file name to search
    :param path (str): path to search within
    :returns (list): returns list of matching paths to filenames or empty list if no matches found. 
    """
    result = []
    for root, _, files in os.walk(path):
        if name in files:
            result.append(Path(os.path.join(root, name)))
    return result


def validate_filepath(filepath):
    filepath = Path(filepath)
    assert filepath.exists()
    return filepath


def get_file_modification_time(filepath, timezone, fmt=None):
    """
    Gets modification time of file.
    
    :param filepath: (patlib.Path or str) path to file
    :timezone: (str) desired timezone in pytz format (e.g. 'US/Central')
    :fmt: optional (str) timestamp format to pass to strftime 
    
    :returns: datetime object
    """
    ts = datetime.datetime.fromtimestamp(Path(filepath).stat().st_mtime, tz=datetime.timezone.utc)
    return timezone_converter(ts, 'UTC', timezone, fmt=fmt)


def append_timestamp_to_filepath(filepath, timestamp, separator='__', with_suffix=None, verbose=True, return_filepath=False):
    """
    Appends timestamp to provided filepath (but before the file extension)
    
    :param filepath: (patlib.Path or str) filepath to directory or file to modify
    :param timestamp: (str or datetime object) timestamp to append
    :param with_suffix: (str) desired alternate file extension to replace existing extension. 
        By default, original extension will be maintained.
    :param verbose: (bool) logs renamed filepath
    :param return_filepath: (bool) returns renamed filepath patlib.Path
    """
    filepath = Path(filepath)  
    filepath_rn = filepath.with_name(f'{filepath.stem}{separator}{timestamp}').with_suffix(filepath.suffix if with_suffix is None else with_suffix)
    filepath.rename(filepath_rn)
    if verbose:
        logging.info(f'File renamed: {filepath_rn}')
    if return_filepath:
        return filepath_rn


"""
Methods for interacting with packages from Allen Institute and Seung lab from Princeton. 

CloudVolume (https://pypi.org/project/cloud-volume/)
CAVE (https://pypi.org/project/caveclient/)
"""

import logging
from cloudvolume import CloudVolume
from caveclient import CAVEclient

logger = logging.getLogger(__name__)

def set_CAVE_client(datastack, ver=None):
    client = CAVEclient(datastack)
    
    if ver is not None:
        client.materialize._version = int(ver)

    logger.info(f'Instantiated CAVE client with version: {client.materialize.version}. Most recent version: {client.materialize.most_recent_version()}')
    
    return client


def get_stats_from_cv_path(cv_path, mip=None):
    """
    Given a cloudvolume path and optional mip (default = all), returns a dict with the following stats from cloudvolume:
        - res: resolution
        - min_pt: min pt of bounding box in voxels
        - max_pt: max pt of bounding box in voxels
        - ctr_pt: center of bounding box in voxels
        - voxel_offset
    
    :param cv_path (str): CloudVolume path
    :param mip (int): the mip to get stats for. default is None.

    :returns: 
        - If mip=None, list of dictionaries for all mips.
        - If mip is specified, dictionaries with stats.  
    """
    def get_stats_for_mip(mip):
        res = cv.mip_resolution(mip)
        min_pt = cv.mip_bounds(mip).minpt
        max_pt = cv.mip_bounds(mip).maxpt
        ctr_pt = ((max_pt - min_pt) / 2) + min_pt
        voxel_offset = cv.mip_voxel_offset(mip)

        return {
                'mip' : mip,
                'res' : res,
                'min_pt' : min_pt,
                'max_pt' : max_pt,
                'ctr_pt' : ctr_pt,
                'voxel_offset' : voxel_offset
            }

    cv = CloudVolume(cv_path, use_https=True, progress=True)

    return get_stats_for_mip(mip) if mip is not None else [get_stats_for_mip(mip) for mip in list(cv.available_mips)]



class MICrONSError(Exception):
    """
    Base class for errors specific to DataJoint internal operation.
    """
    def suggest(self, *args):
        """
        regenerate the exception with additional arguments
        :param args: addition arguments
        :return: a new exception of the same type with the additional arguments
        """
        return self.__class__(*(self.args + args))

class VersionError(MICrONSError):
    """
    Incorrect version.
    """


"""
Classes and methods for configuring microns packages.
"""
import datajoint_plus as djp

class SchemaConfig:
    """
    DataJoint schema configuration object.
    """
    def __init__(self, module_name:str, schema_name:str, externals:dict=None, adapters:dict=None):
        """
        Initializes DataJoint configuration object.
        
        :param module_name: name of Python module containing DataJoint tables
        :param schema_name: name of DataJoint schema object
        :param externals: dictionary of externals folders to map to schema
        :param adapters: adapters to add to schema context
        """
        self._module_name = module_name
        self._schema_name = schema_name
        self._externals = externals
        self._adapters = adapters
    
    @property
    def module_name(self):
        return self._module_name
    
    @property
    def schema_name(self):
        return self._schema_name
    
    @property
    def externals(self):
        return self._externals
    
    @property
    def adapters(self):
        return self._adapters
    
    def register_externals(self):
        if self.externals is not None:
            djp.register_externals(self.externals)
    
    def register_adapters(self, context=None):
        if self.adapters is not None:
            djp.add_objects(self.adapters, context=context)



"""Mesh utilities."""

import numpy as np

def index_unique_rows(full_coordinate_array):
    """
    Separates an array of nested coordinate rows into an array of unique rows and and index array.
    """
    vertices, flat_idx = np.unique(full_coordinate_array.reshape(-1, full_coordinate_array.shape[-1]), axis=0, return_inverse=True)
    return vertices, flat_idx.reshape(-1, full_coordinate_array.shape[-2])

def get_midpoints(edges):
    return (edges[:, 0] + edges[:, 1]) / 2

def get_thresholded_bbox(vertices, threshold):
     return np.array((vertices.min(0), vertices.max(0))) + np.array((-threshold, threshold))[:, None] 

def bbox_point_containment(points, bbox):
    """
    Returns a mask of which points are within the boundary described by the bbox at all.
    """
    return np.logical_and.reduce((
        points[..., 0] >= bbox[0, 0],
        points[..., 0] <= bbox[1, 0],
        points[..., 1] >= bbox[0, 1],
        points[..., 1] <= bbox[1, 1],
        points[..., 2] >= bbox[0, 2],
        points[..., 2] <= bbox[1, 2],
    ), axis=0)





"""
Utils for DataJoint adapters
"""

import datajoint as dj
import numpy as np
import h5py
from pathlib import Path
from collections import namedtuple
import pickle
from datetime import datetime
import logging

from .filepath_utils import validate_filepath


class Adapter(dj.AttributeAdapter):
    attribute_type = ''

    def __init__(self, attribute_type):
        self.attribute_type = attribute_type
        super().__init__()


class FilePathAdapter(Adapter):    
    def put(self, filepath):
        return validate_filepath(filepath)

    def get(self, filepath):
        return validate_filepath(filepath)


class PickleAdapter(Adapter):
    def put(self, object):
        return pickle.dumps(object)

    def get(self, pickled):
        return pickle.loads(pickled)


class MeshAdapter(FilePathAdapter):
    def get(self, filepath):
        filepath = super().get(filepath)
        return adapt_mesh_hdf5(filepath, filepath_has_timestamp=True)


def adapt_mesh_hdf5(filepath, parse_filepath_stem=True, filepath_has_timestamp=False, separator='__', timestamp_fmt="%Y-%m-%d_%H:%M:%S", return_type='namedtuple', as_lengths=False):
    """
    Reads from a mesh hdf5 and returns vertices, faces and additional information in the form of a namedtuple or optionally
        as a dictionary, or optionally as separate variables.
        
    :param filepath: File path pointing to the hdf5 mesh file. 
    :param parse_filepath_stem: (bool) 
        If True: Attempts to parse filepath stem
        If False: Skips parsing of filepath stem
    :param filepath_has_timestamp: (bool) Toggles format of expected filename stem to parse.
        If True: expects format '<segment_id><separator><timestamp>'.h5
        If False: expects format '<segment_id>'.h5
    :param timestamp_separator: (str) 
    :param timestamp_fmt:
    :param return_type: Options = {
        namedtuple = return a namedtuple with the following fields {
            vertices = vertex array
            faces = face array
            segment_id = segment id of mesh if parsed else np.nan
            timestamp = timestamp mesh was computed if parsed else ''
            filepath = filepath of mesh
        }
        dict = return a dictionary with keys as in namedtuple
        separate: returns separate variables in the following order {
            vertex array 
            face array
            dictionary with keys segment_id, timestamp, filepath as in namedtuple
            }
    :param as_lengths: Overrides return_type and instead returns:
            Length of the vertex array
            Length of face array
            dictionary with keys segment_id, timestamp, filepath as in namedtuple
        This is done without pulling the mesh into memory, which makes it far more space and time efficient.
    }
    """
    Mesh = namedtuple('Mesh', ['vertices', 'faces', 'segment_id', 'timestamp', 'filepath'])
    filepath = Path(filepath)
    
    # try to parse filepath
    info_dict = {'filepath': filepath}
    defaults = {'segment_id': np.nan, 'timestamp': ''}
    if parse_filepath_stem:
        try:
            if not filepath_has_timestamp:
                segment_id = filepath.stem
                info_dict.update({**{'segment_id': int(segment_id), 'timestamp': ''}})
            else:
                segment_id, timestamp = filepath.stem.split(separator)
                timestamp = datetime.strptime(timestamp, timestamp_fmt)
                info_dict.update({**{'segment_id': int(segment_id), 'timestamp': timestamp}})
        except:
            info_dict.update({**defaults})
            logging.warning('Could not parse mesh filepath.')
    else:
        info_dict.update({**defaults})

    # Load the mesh data
    with h5py.File(filepath, 'r') as f:
        if as_lengths:
            n_vertices = f['vertices'].shape[0]
            n_faces = int(f['faces'].shape[0] / 3)
            return n_vertices, n_faces, info_dict
        
        vertices = f['vertices'][()].astype(np.float64)
        faces = f['faces'][()].reshape(-1, 3).astype(np.uint32)
    
    # Return options
    return_dict = dict(
                vertices=vertices,
                faces=faces,
                **info_dict
                )
    if return_type == 'namedtuple':
        return Mesh(**return_dict)
    elif return_type == 'dict':
        return return_dict
    elif return_type == 'separate':
        return vertices, faces, info_dict
    else:
        raise TypeError(f'return_type does not accept {return_type} argument')


#!/usr/bin/env python
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'microns_utils', 'version.py')) as f:
    exec(f.read())

with open(path.join(here, 'requirements.txt')) as f:
    requirements = f.read().split()

setup(
    name='microns-utils',
    version=__version__,
    description='utilities for MICrONS',
    author='Stelios Papadopoulos, Christos Papadopoulos',
    author_email='spapadop@bcm.edu, cpapadop@bcm.edu',
    packages=find_packages(exclude=[]),
    install_requires=requirements
)